package industries.sunshine.planningpoker

import cats.effect._
import cats.syntax.all._
import cats.data.Kleisli
import cats.data.OptionT
import org.http4s.Request
import industries.sunshine.planningpoker.common.Models.PlayerID
import java.util.UUID
import industries.sunshine.planningpoker.common.Models.{RoomID, Room}
import org.http4s.ResponseCookie
import cats.data.EitherT
import scala.util.Random

trait Auth[F[_]] {

  /** for middleware that converts sessionId into PlayerId
    */
  def authUser: Kleisli[[A] =>> cats.data.OptionT[F, A], Request[F], (PlayerID, RoomID)]

  /** Get sessionId for accessing a room
    * @param roomPassword
    *   \- requirement to get access to the room
    *
    * check that room exists, password is valid call to add user to the players create session
    * mapping and return cookie
    */
  def joinRoom(
      roomName: String,
      roomPassword: String,
      nickName: String,
      nickPassword: String
  ): F[Either[String, ResponseCookie]]

  def deleteSession(
      sessionId: Long
  ): F[Unit]

}

object Auth {
  def pureBadStub = new Auth[IO] {
    override def authUser =
      Kleisli((req: Request[IO]) =>
        OptionT.liftF(
          IO(println(s"authUser: $req")) >> IO(
            PlayerID(14) -> RoomID("testroom")
          )
        )
      )
    override def joinRoom(
        roomName: String,
        roomPassword: String,
        nickName: String,
        nickPassword: String
    ) =
      IO(
        println(
          s"> access room for $roomName $roomPassword $nickName, to return stub 111"
        )
      ) >> IO.pure(Right(ResponseCookie("authcookieName", "1")))

    override def deleteSession(sessionId: Long): IO[Unit] =
      IO(s"got request to leave for $sessionId")
  }

  type SessionsMap = Map[Long, (RoomID, PlayerID)]

  class SimpleAuth[F[_]: Sync](
      sessions: Ref[F, SessionsMap],
      roomService: RoomService[F]
  ) extends Auth[F] {

    val authcookieName = "authcookie"

    override def joinRoom(
        roomName: String,
        roomPassword: String,
        nickName: String,
        nickPassword: String
    ): F[Either[String, ResponseCookie]] = {
      // TODO check for existing session for same room
      // and do i want to logout if existing session for another room? ugh
      val roomId = RoomID(roomName)
      val result = for {
        playerId <- EitherT(
          roomService.joinRoom(roomId, nickName, nickPassword, roomPassword)
        )
          .leftMap(_.toString())
        newSessionId = Random.nextLong()
        _ <- EitherT.liftF(sessions.update(_.updated(newSessionId, (roomId, playerId))))
      } yield ResponseCookie(
        name = authcookieName,
        content = newSessionId.toString(),
        secure = true
      )

      result.value
    }

    override def authUser
        : Kleisli[[A] =>> cats.data.OptionT[F, A], Request[F], (PlayerID, RoomID)] = {
      Kleisli { (request: Request[F]) =>
        OptionT(sessions.get.map { sessionsMap =>
          for {
            authcookie <- request.cookies.find(_.name == authcookieName)
            sessionId <- authcookie.content.toLongOption
            (roomId, playerId) <- sessionsMap.get(sessionId)
          } yield (playerId, roomId)
        })
      }
    }

    override def deleteSession(sessionId: Long): F[Unit] = {
      sessions.update(_.removed(sessionId))
    }
  }

  def make[F[_]: Sync](roomsService: RoomService[F]): F[Auth[F]] =
    for {
      sessionsMap <- Ref.of[F, SessionsMap](
        Map(1L -> (RoomID("testroom"), PlayerID(1L)))
      )
    } yield new SimpleAuth(sessionsMap, roomsService)
}
